function Color.rainbow()
    local t = Application:time()
    local r,g,b = 1,1,1
    local h = (t % 5) / 5 * 6
    local i = math.floor(h)
    local f = h - i
    local q = 1 - f

    if i % 6 == 0 then
        r, g, b = 1, f, 0
    elseif i % 6 == 1 then
        r, g, b = q, 1, 0
    elseif i % 6 == 2 then
        r, g, b = 0, 1, f
    elseif i % 6 == 3 then
        r, g, b = 0, q, 1
    elseif i % 6 == 4 then
        r, g, b = f, 0, 1
    elseif i % 6 == 5 then
        r, g, b = 1, 0, q
    end

    return r,g,b
end

local function Pulse(speed)
    local time = Application:time()
    local opacity = math.abs(math.sin(time * (speed or 1)))
    return opacity
end

local function custom_colors_con(value)
    local c = {
        pastel_pink = {255, 161, 220},
        purple = {128, 0, 128},
        aqua = {0, 255, 221},
        strawb = {251, 41, 65},
        orange = {255, 85, 0},
        red = {255, 0, 0},
        navy = {56, 63, 255},
        pink = {255, 0, 160},
        lilac = {223, 168, 255},
        black = {0, 0, 0},
        blue_violet = {169, 48, 255},
        white = {255, 255, 255},
        green = {0, 255, 0},
        yellow = {255, 255, 0},
        vanilla_color = {153, 51, 51}
    }

    local convert = {
        "pastel_pink",
        "purple",
        "aqua",
        "strawb",
        "orange",
        "red",
        "navy",
        "pink",
        "lilac",
        "black",
        "blue_violet",
        "white",
        "rainbow",
        "green",
        "yellow",
        "vanilla_color"
    }

    if not value then
        return c.red[1]/255, c.red[2]/255, c.red[3]/255
    end

    if value == 13 then
        return Color.rainbow()
    end

    return c[convert[value]][1]/255, c[convert[value]][2]/255, c[convert[value]][3]/255
end

local function ColorUnit(unit)
    local ColorUnitOptions = {
        medic = Contours_Opt._data.medic_color,
        shield = Contours_Opt._data.shield_color,
        sniper = Contours_Opt._data.sniper_color,
        taser = Contours_Opt._data.taser_color,
        spooc = Contours_Opt._data.spooc_color,
        tank = Contours_Opt._data.tank_color,
        tank_hw = Contours_Opt._data.tank_hw_color,
        phalanx_vip = Contours_Opt._data.shield_captain_color,
        phalanx_minion =  Contours_Opt._data.phalanx_vip_color,
        tank_medic = Contours_Opt._data.tank_medic_color,
        tank_mini = Contours_Opt._data.tank_mini_color,
        marshal_shield = Contours_Opt._data.shield_color,
        marshal_marksman = Contours_Opt._data.sniper_color,

        --Bosses
        hector_boss = Contours_Opt._data.boss_color,
        hector_boss_no_armor = Contours_Opt._data.boss_color,
        mobster_boss = Contours_Opt._data.boss_color,
        biker_boss = Contours_Opt._data.boss_color,
        drug_lord_boss = Contours_Opt._data.boss_color,
        drug_lord_boss_stealth = Contours_Opt._data.boss_color,
        triad_boss = Contours_Opt._data.boss_color,
        triad_boss_no_armor = Contours_Opt._data.boss_color,
        deep_boss = Contours_Opt._data.boss_color,

        --Resmod units
        summers = Contours_Opt._data.summers_color,
        boom_summers = Contours_Opt._data.boom_summers_color,
        taser_summers = Contours_Opt._data.taser_summers_color,
        medic_summers = Contours_Opt._data.medic_summers_color,
        spring = Contours_Opt._data.spring_color,
        headless_hatman = Contours_Opt._data.headless_hatman_color,
        tank_hw_black = Contours_Opt._data.tank_hw_color,
        phalanx_minion_assault = Contours_Opt._data.plananx_minion_assault_color,
        taser_titan = Contours_Opt._data.taser_titan_color,
        taser_titan_reaper = Contours_Opt._data.taser_titan_reaper_color,
        heavy_swat_sniper = Contours_Opt._data.heavy_swat_sniper_color,
        tank_titan = Contours_Opt._data.tank_titan_color,
        tank_titan_assault = Contours_Opt._data.tank_titan_assault_color,
        tank_biker = Contours_Opt._data.tank_color,
        hrt_titan = Contours_Opt._data.hrt_titan_color,
        omnia_lpf = Contours_Opt._data.omnia_lpf_color,
        enforcer = Contours_Opt._data.enforcer_assault_color,
        enforcer_assault = Contours_Opt._data.enforcer_assault_color,
        boom = Contours_Opt._data.boom_color,
        city_swat_titan = Contours_Opt._data.city_swat_titan_color,
        city_swat_titan_assault	= Contours_Opt._data.city_swat_titan_assault_color,
        weekend_dmr = Contours_Opt._data.heavy_swat_sniper_color,
        weekend_lmg = Contours_Opt._data.city_swat_titan_color,
        fbi_vet = Contours_Opt._data.fbi_vet_color,
        vetlod = Contours_Opt._data.fbi_vet_color
    }

    return ColorUnitOptions[unit]
end

local function upd_rgb_data(special, unit)
    local color_v = Contours_Opt._data.Contours_Opt_color

    if special then
        return custom_colors_con(ColorUnit(unit))
    end

    return custom_colors_con(color_v)
end

local WantedContours = {
    "mark_enemy",
    "mark_enemy_damage_bonus",
    "mark_enemy_damage_bonus_distance",
    "mark_unit_dangerous",
    "mark_unit_dangerous_damage_bonus",
    "mark_unit_dangerous_damage_bonus_distance"
}

local EnemyTypes = {
    RegularUnits = {
        "security",
        "security_mex",
        "security_mex_no_pager",
        "gangster",
        "biker",
        "bolivian",
        "bolivian_indoors",
        "bolivian_indoors_mex",
        "triad",
        "gensec",
        "murky",
        "cop",
        "cop_scared",
        "fbi",
        "swat",
        "heavy_swat",
        "fbi_swat",
        "fbi_heavy_swat",
        "city_swat",
        --Resmod units
        "hrt",
        "city_swat_guard",
        "gensec_guard",
        "biker_guard",
        "weekend",
        "weekend_guard",
        "weekend_elite_guard",
        "spooc_gangster"
    },
    SpecialUnits = {
        "shield",
        "sniper",
        "medic",
        "taser",
        "spooc",
        "tank",
        "tank_hw",
        "phalanx_vip",
        "phalanx_minion",
        "tank_medic",
        "tank_mini",
        "marshal_shield",
		"marshal_marksman",
        --Resmod captains, specials and elites
        "summers",
        "boom_summers",
        "taser_summers",
        "medic_summers",
        "spring",
        "headless_hatman",
        "tank_hw_black",
        "phalanx_minion_assault",
        "taser_titan",
        "taser_titan_reaper",
        "heavy_swat_sniper",
        "tank_titan",
        "tank_titan_assault",
        "tank_biker",
        "hrt_titan",
        "omnia_lpf",
        "enforcer",
        "enforcer_assault",
        "boom",
        "city_swat_titan",
        "city_swat_titan_assault",
        "weekend_dmr",
        "weekend_lmg",
        "fbi_vet",
        "vetlod"
    },
    Bosses = {
        "hector_boss",
        "hector_boss_no_armor",
        "biker_boss",
        "drug_lord_boss",
        "drug_lord_boss_stealth",
        "triad_boss",
        "triad_boss_no_armor",
        "deep_boss"
    }
}

local function ContourBoss(e)
    local enemy_u = e and e:base() and e:base()._tweak_table
    for _, id_s in ipairs(EnemyTypes.Bosses) do
        if enemy_u == id_s then
            local r,g,b = upd_rgb_data(true, id_s)
            return Vector3(r,g,b), id_s
        end
    end
end

local function ContourSpec(e)
    local enemy_u = e and e:base() and e:base()._tweak_table
    for _, id in ipairs(EnemyTypes.SpecialUnits) do
        if enemy_u == id then
            local r,g,b = upd_rgb_data(true, id)
            return Vector3(r,g,b), id
        end
    end
end

local function ContourRegular(e)
    local enemy_u = e and e:base() and e:base()._tweak_table
    for _, id_r in ipairs(EnemyTypes.RegularUnits) do
        if enemy_u == id_r then
            local r,g,b = upd_rgb_data()
            return Vector3(r,g,b), id_r
        end
    end
end

local function CheckPulseOpt()
    local PulseRegular = Contours_Opt._data.con_mode_r == 2
    local PulseSpecials = Contours_Opt._data.con_mode_s == 2
    local PulseBoss = Contours_Opt._data.con_mode_b == 2
    local PulseChk = PulseBoss or PulseRegular or PulseSpecials

    return PulseChk
end

local function PulseChangeColor(unique, c, id, e)
    local PulseRegular = Contours_Opt._data.con_mode_r
    local PulseSpecials = Contours_Opt._data.con_mode_s
    local PulseBoss = Contours_Opt._data.con_mode_b
    local Regular, i = ContourRegular(e)
    local Boss, i2 = ContourBoss(e)
    local Spec, i3 = ContourSpec(e)
    local r2,g2,b2 = upd_rgb_data()
    local pdataorig = {r2, g2, b2}
    local PulseSpd = Contours_Opt._data.con_spd
    local Color = Vector3(0,0,0)

    if unique then
        if Regular then
            if PulseRegular == 2 then
                local r,g,b = Regular[1],Regular[2],Regular[3]
                r = r - Pulse(PulseSpd)
                g = g - Pulse(PulseSpd)
                b = b - Pulse(PulseSpd)
                Color = Vector3(r,g,b)
                c:change_color(id, Color)
            else
                c:change_color(id, Regular)
            end
        end

        if Spec then
            if PulseSpecials == 2 then
                local r,g,b = Spec[1],Spec[2],Spec[3]
                r = r - Pulse(PulseSpd)
                g = g - Pulse(PulseSpd)
                b = b - Pulse(PulseSpd)
                Color = Vector3(r,g,b)
                c:change_color(id, Color)
            else
                c:change_color(id, Spec)
            end
        end

        if Boss then
            if PulseBoss == 2 then
                local r,g,b = Boss[1],Boss[2],Boss[3]
                r = r - Pulse(PulseSpd)
                g = g - Pulse(PulseSpd)
                b = b - Pulse(PulseSpd)
                Color = Vector3(r,g,b)
                c:change_color(id, Color)
            else
                c:change_color(id, Boss)
            end
        end
    else
        if PulseRegular == 2 or PulseBoss == 2 or PulseSpecials == 2 then
            pdataorig[1] = pdataorig[1] - Pulse(PulseSpd)
            pdataorig[2] = pdataorig[2] - Pulse(PulseSpd)
            pdataorig[3] = pdataorig[3] - Pulse(PulseSpd)
        end

        if Regular then
            if PulseRegular == 2 then
                c:change_color(id, Vector3(pdataorig[1],pdataorig[2],pdataorig[3]))
            else
                c:change_color(id, Vector3(r2,g2,b2))
            end
        end

        if Spec then
            if PulseSpecials == 2 then
                c:change_color(id, Vector3(pdataorig[1],pdataorig[2],pdataorig[3]))
            else
                c:change_color(id, Vector3(r2,g2,b2))
            end
        end

        if Boss then
            if PulseBoss == 2 then
                c:change_color(id, Vector3(pdataorig[1],pdataorig[2],pdataorig[3]))
            else
                c:change_color(id, Vector3(r2,g2,b2))
            end
        end
    end
end

local function ContoursPlusChangeColor(unique, c, id, e)
    local Regular, i = ContourRegular(e)
    local Boss, i2 = ContourBoss(e)
    local Spec, i3 = ContourSpec(e)
    local r2,g2,b2 = upd_rgb_data()

    if unique then
        if Regular then
            c:change_color(id, Regular)
        end

        if Spec then
            c:change_color(id, Spec)
        end

        if Boss then
            c:change_color(id, Boss)
        end
    else
        c:change_color(id, Vector3(r2,g2,b2))
    end
end

local function init_color_change(contourExt, enemy)
    for _, mark_id in ipairs(WantedContours) do
        if not CheckPulseOpt then
            ContoursPlusChangeColor(Contours_Opt._data.unique_contours, contourExt, mark_id, enemy)
        elseif CheckPulseOpt then
            PulseChangeColor(Contours_Opt._data.unique_contours, contourExt, mark_id, enemy)
        end
        contourExt:update_materials()
    end
end

function AddContour(contourExt, enemy)
    init_color_change(contourExt, enemy)
end

local store_enemies = {}

function FetchEnemies()
    for _, enemy in pairs(store_enemies) do
        local contourExt = enemy and alive(enemy) and enemy:contour()
        if contourExt then
            AddContour(contourExt, enemy)
        end
    end
end

Hooks:PreHook(ContourExt, "add", "add_addonsad", function(self, type, sync, multiplier, override_color, add_as_child)
    for _, mark_id in ipairs(WantedContours) do
        if type == mark_id and alive(self._unit) then
            for _, stored_enemy in pairs(store_enemies) do
                if stored_enemy == self._unit then
                    return
                end
            end
            table.insert(store_enemies, self._unit)
        end
    end
end)

Hooks:PreHook(ContourExt, "remove", "remove_addonsad", function(self, type, sync)
    for _, mark_id in ipairs(WantedContours) do
        if type == mark_id and alive(self._unit) then
            for i, stored_enemy in pairs(store_enemies) do
                if stored_enemy == self._unit then
                    table.remove(store_enemies, i)
                end
            end
        end
    end
end)

local function FetchCallback()
    if Contours_Opt._data.enabled_con then
        FetchEnemies()
    end
end

Hooks:Add("GameSetupUpdate", "init_ContoursUpdate", FetchCallback)